from strictyaml import Map, Regex, Seq, Str, Int, Optional, Seq, MapPattern, Enum, Bool

limits_schema = Map(
    {
        "time": Int(),
        Optional("memory", default=8192): Int(),
        Optional("output"): Int(),
        Optional("cores"): Str(),
    }
)

module_schema = Regex(r"\.?\w+(\.\w+)*")

step_schema = Seq(Map({"step": module_schema}))

schema = Map(
    {
        "title": Str(),
        Optional("description"): Str(),
        "limits": limits_schema,
        "steps": Map({"run": step_schema, Optional("compile"): step_schema}),
        "tasks": MapPattern(
            Str(), Map({"type": Enum(["folder"]), Optional("path"): Str()})
        ),
        "tools": MapPattern(Str(), module_schema),
        "parameters": MapPattern(Str(), MapPattern(Str(), Str())),
    }
)



class Runner:
    def __init__(self, config):
        self.config = config

    def run(self):
        pass


class Step:
    def run(self, context):
        pass


class Tool:
    name = "Base Tool"
    REQUIRED_PATHS = []

    def setup(self):
        pass

    def version(self):
        return "1.0.0"

    def pre_run(self, context):
        pass

    def cmdline(self, context):
        pass

    def post_run(self, context):
        pass

    def teardown(self):
        pass



import subprocess
import functools
import operator

from reprobench.core.bases import Step
from reprobench.utils import find_executable, silent_run

# Not working yet
class IsolateExecutor(Step):
    def __init__(self):
        self.executable = find_executable("isolate")
    
    def run(self, context):
        tool = context['tool']
        tool.pre_run(context['task'], context['parameters'])

        output = subprocess.check_output(
        # print(
            [ self.executable, "--run", "-e" ] +
            functools.reduce(operator.iconcat, [ ["-d", f"{path}:rw"] for path in tool.REQUIRED_PATHS ]) +
            [ "--" ] +
            tool.cmdline(context['task'], context['parameters'])
        )
        return output



import pynisher
import subprocess
from loguru import logger
from pathlib import Path

from reprobench.core.bases import Step
from reprobench.core.db import db, Run, RunStatistic


class PynisherExecutor(Step):
    def run(self, context):
        tool = context["tool"]
        limits = context["limits"]
        tool.pre_run(context)

        cwd = context["run"].directory
        out_file = (Path(cwd) / "run.out").open("wb")
        err_file = (Path(cwd) / "run.err").open("wb")

        context["run"].status = Run.RUNNING
        context["run"].save()

        def run_tool():
            subprocess.run(
                tool.cmdline(context), cwd=cwd, stdout=out_file, stderr=err_file
            )

        logger.debug(f"Running {cwd}")

        fun = pynisher.enforce_limits(
            cpu_time_in_s=limits["time"], mem_in_mb=limits["memory"]
        )(run_tool)

        fun()

        logger.debug(f"Finished {cwd}")

        context["run"].status = Run.DONE
        context["run"].verdict = Run.SUCCESS
        context["run"].save()
        RunStatistic.create(
            run=context["run"], key=RunStatistic.WALL_TIME, value=fun.wall_clock_time
        )
        RunStatistic.create(
            run=context["run"],
            key=RunStatistic.CPU_TIME,
            value=fun.resources_function[0] + fun.resources_function[1],
        )  # utime
        RunStatistic.create(
            run=context["run"],
            key=RunStatistic.MEM_USAGE,
            value=fun.resources_function[2],
        )  # maxrss

        tool.post_run(context)




import subprocess
from loguru import logger
from pathlib import Path
from psmon.main import run

from reprobench.core.bases import Step
from reprobench.core.db import db, Run, RunStatistic


class PsmonExecutor(Step):
    def run(self, context):
        tool = context["tool"]
        limits = context["limits"]
        tool.pre_run(context)

        cwd = context["run"].directory
        out_file = (Path(cwd) / "run.out").open("wb")
        err_file = (Path(cwd) / "run.err").open("wb")

        context["run"].status = Run.RUNNING
        context["run"].save()

        cmd = tool.cmdline(context)
        logger.debug(f"Running {cwd}")
        logger.trace(cmd)

        stats = run(
            cmd,
            cwd=cwd,
            stdout=out_file,
            stderr=err_file,
            cpu_time_limit=limits["time"],
            wall_time_limit=limits["time"] + 15,
            memory_limit=limits["memory"],
            freq=15,
        )

        logger.debug(f"Finished {cwd}")
        logger.debug(stats)

        context["run"].status = Run.DONE
        context["run"].return_code = stats["return_code"]

        if stats["error"] == TimeoutError:
            context["run"].verdict = Run.TIMEOUT
        elif stats["error"] == MemoryError:
            context["run"].verdict = Run.MEMOUT
        elif stats["error"] or stats["return_code"] != 0:
            context["run"].verdict = Run.RUNTIME_ERR
        else:
            context["run"].verdict = Run.SUCCESS

        context["run"].save()

        RunStatistic.create(
            run=context["run"], key=RunStatistic.WALL_TIME, value=stats["wall_time"]
        )
        RunStatistic.create(
            run=context["run"], key=RunStatistic.CPU_TIME, value=stats["cpu_time"]
        )
        RunStatistic.create(
            run=context["run"], key=RunStatistic.MEM_USAGE, value=stats["max_memory"]
        )

        tool.post_run(context)




import os
import signal
import itertools
import time
import atexit
import subprocess
from string import Template
from tqdm import tqdm
from loguru import logger
from multiprocessing.pool import Pool
from pathlib import Path
from datetime import datetime
from playhouse.apsw_ext import APSWDatabase
from reprobench.core.bases import Runner
from reprobench.core.db import db, db_bootstrap, Run, Tool, ParameterCategory, Task
from reprobench.utils import import_class

DIR = os.path.dirname(os.path.realpath(__file__))


class SlurmRunner(Runner):
    def __init__(
        self,
        config,
        config_path,
        python_path,
        output_dir="./output",
        resume=False,
        teardown=False,
    ):
        self.config = config
        self.config_path = config_path
        self.output_dir = output_dir
        self.python_path = python_path
        self.resume = resume
        self.teardown = teardown
        self.queue = []

    def setup(self):
        atexit.register(self.exit)

        self.db_path = Path(self.output_dir) / f"{self.config['title']}.benchmark.db"
        db_created = Path(self.db_path).is_file()

        if db_created and not self.resume:
            logger.error(
                "It seems that a previous runs exist at the output directory. Please use --resume to resume runs."
            )
            exit(1)

        Path(self.output_dir).mkdir(parents=True, exist_ok=True)
        logger.debug(f"Creating Database: {self.db_path}")
        self.database = APSWDatabase(str(self.db_path))
        db.initialize(self.database)

        if not db_created:
            logger.info("Bootstrapping db...")
            db_bootstrap(self.config)
            logger.info("Initializing runs...")
            self.init_runs()

    def create_working_directory(
        self, tool_name, parameter_category, task_category, filename
    ):
        path = (
            Path(self.output_dir)
            / tool_name
            / parameter_category
            / task_category
            / filename
        )
        path.mkdir(parents=True, exist_ok=True)
        return path

    def exit(self):
        pass

    def init_runs(self):
        for tool_name, tool_module in self.config["tools"].items():
            for (parameter_category_name, (task_category, task)) in itertools.product(
                self.config["parameters"], self.config["tasks"].items()
            ):
                # only folder task type for now
                assert task["type"] == "folder"

                files = Path().glob(task["path"])
                for file in files:
                    context = self.config.copy()
                    directory = self.create_working_directory(
                        tool_name, parameter_category_name, task_category, file.name
                    )

                    tool = Tool.get(Tool.module == tool_module)
                    parameter_category = ParameterCategory.get(
                        ParameterCategory.title == parameter_category_name
                    )
                    task = Task.get(Task.path == str(file))

                    run = Run.create(
                        tool=tool,
                        task=task,
                        parameter_category=parameter_category,
                        directory=directory,
                    )

                    self.queue.append(run.id)

    def run(self):
        if not self.teardown:
            self.setup()
            logger.debug("Running setup on all tools...")
            tools = []
            for tool_module in self.config["tools"].values():
                ToolClass = import_class(tool_module)
                tool_instance = ToolClass()
                tool_instance.setup()
                tools.append(tool_instance)
            logger.debug("Generating template")

            with open(Path(DIR) / "./slurm.job.tpl") as tpl:
                template = Template(tpl.read())
                job_str = template.safe_substitute(
                    python_path=self.python_path,
                    config_path=self.config_path,
                    db_path=self.db_path,
                )

            slurm_job_path = Path(self.output_dir) / "slurm.job"
            with open(slurm_job_path, "w") as job:
                job.write(job_str)

            logger.info("Submitting job array to SLURM...")
            sbatch_cmd = [
                "sbatch",
                "-a",
                f"1-{len(self.queue)}",
                str(slurm_job_path.resolve()),
            ]
            logger.debug(" ".join(sbatch_cmd))
            subprocess.run(sbatch_cmd)
        else:
            logger.debug("Running teardown on all tools...")
            for tool_module in self.config["tools"].values():
                ToolClass = import_class(tool_module)
                tool_instance = ToolClass()
                tool_instance.teardown()



from pathlib import Path
from datetime import datetime
from peewee import Proxy, Model
from playhouse.apsw_ext import (
    DateTimeField,
    CharField,
    ForeignKeyField,
    IntegerField,
    BooleanField,
)

db = Proxy()


class BaseModel(Model):
    created_at = DateTimeField(default=datetime.now)

    class Meta:
        database = db


class Limit(BaseModel):
    type = CharField(max_length=32, unique=True)
    value = CharField()


class TaskCategory(BaseModel):
    title = CharField()


class Task(BaseModel):
    category = ForeignKeyField(TaskCategory, backref="tasks", on_delete="cascade")
    path = CharField()


class Tool(BaseModel):
    module = CharField(unique=True)
    name = CharField()
    version = CharField(null=True)


class ParameterCategory(BaseModel):
    title = CharField()


class Parameter(BaseModel):
    category = ForeignKeyField(
        ParameterCategory, backref="parameters", on_delete="cascade"
    )
    key = CharField()
    value = CharField()

    class Meta:
        indexes = ((("category", "key"), True),)


class Run(BaseModel):
    FAILED = -2
    CANCELED = -1
    PENDING = 0
    SUBMITTED = 1
    RUNNING = 2
    DONE = 3

    STATUS_CHOICES = (
        (FAILED, "Failed"),
        (CANCELED, "Canceled"),
        (PENDING, "Pending"),
        (SUBMITTED, "Submitted"),
        (RUNNING, "Running"),
        (DONE, "Done"),
    )

    TIMEOUT = "TLE"
    MEMOUT = "MEM"
    RUNTIME_ERR = "RTE"
    OUTPUT_LIMIT = "OLE"
    SUCCESS = "OK"

    VERDICT_CHOICES = (
        (TIMEOUT, "Time Limit Exceeded"),
        (MEMOUT, "Memory Limit Exceeded"),
        (RUNTIME_ERR, "Runtime Error"),
        (OUTPUT_LIMIT, "Output Limit Exceeded"),
        (SUCCESS, "Run Successfully"),
    )

    tool = ForeignKeyField(Tool, backref="runs", on_delete="cascade")
    parameter_category = ForeignKeyField(
        ParameterCategory, backref="runs", on_delete="cascade"
    )
    task = ForeignKeyField(Task, backref="runs", on_delete="cascade")
    status = IntegerField(choices=STATUS_CHOICES, default=PENDING)
    verdict = CharField(choices=VERDICT_CHOICES, max_length=3, null=True)
    directory = CharField(null=True)
    valid = BooleanField(null=True)
    return_signal = IntegerField(null=True)
    return_code = IntegerField(null=True)

    class Meta:
        only_save_dirty = True


class RunStatistic(BaseModel):
    CPU_TIME = "cpu"
    WALL_TIME = "wall"
    MEM_USAGE = "mem"

    KEY_CHOICES = (
        (CPU_TIME, "CPU Time (s)"),
        (WALL_TIME, "Wall Clock Time (s)"),
        (MEM_USAGE, "Max Memory Usage (KiB)"),
    )

    run = ForeignKeyField(Run, backref="statistics", on_delete="cascade")
    key = CharField(choices=KEY_CHOICES)
    value = CharField()


MODELS = [
    Limit,
    TaskCategory,
    Task,
    ParameterCategory,
    Parameter,
    Run,
    RunStatistic,
    Tool,
]


def db_bootstrap(config):
    db.connect()
    db.create_tables(MODELS)

    Limit.insert_many(
        [{"type": key, "value": value} for (key, value) in config["limits"].items()]
    ).execute()

    Tool.insert_many(
        [{"name": name, "module": module} for (name, module) in config["tools"].items()]
    ).execute()

    for (category, parameters) in config["parameters"].items():
        parameter_category = ParameterCategory.create(title=category)
        for (key, value) in parameters.items():
            Parameter.create(category=parameter_category, key=key, value=value)

    for (category, task) in config["tasks"].items():
        task_category = TaskCategory.create(title=category)
        assert task["type"] == "folder"
        for file in Path().glob(task["path"]):
            Task.create(category=task_category, path=str(file))



import logging
import importlib
import subprocess
from reprobench.core.exceptions import ExecutableNotFoundError
from shutil import which

log = logging.getLogger(__name__)


def find_executable(executable):
    path = which(executable)
    if path is None:
        raise ExecutableNotFoundError
    return path


def silent_run(command):
    log.debug(f"Running: {command}")
    return subprocess.run(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def import_class(path):
    module_path, tail = ".".join(path.split(".")[:-1]), path.split(".")[-1]
    module = importlib.import_module(module_path)
    return getattr(module, tail)
